import pandas as pds
import numpy as np
import copy
#import meta

from pysat import DataFrame, Series, datetime, Panel

def set_data_dir(path=None):
    """
    set the top level directory pysat uses to look for data.
    """
    import os
    import pysat
    if os.path.isdir(path):
        with open(os.path.join(os.getenv('HOME'), '.pysat','data_path.txt'),'w') as f:
            f.write(path)
            pysat.data_dir = path
            reload(pysat._files)
    else:
        raise ValueError('Path does not lead to a valid directory.')
        

def load_netcdf3(fnames=None, strict_meta=False, index_label=None,
                    unix_time=False, **kwargs):
                    
    import netCDF4
    import string
    from pysat import meta

    
    if fnames is None:
        raise ValueError("Must supply a list of filenames")
    if not hasattr(fnames, '__iter__'):
        fnames = [fnames]

    saved_mdata = None
    running_idx = 0
    running_store=[]
    two_d_keys = []; two_d_dims = [];
    for fname in fnames:
        with netCDF4.Dataset(fname, mode='r', format='NETCDF3_64BIT') as data:
            # build up dictionary with all ncattrs
            # and add those attributes to a pysat meta object
            ncattrsList = data.ncattrs()
            mdata = meta.Meta()
            for d in ncattrsList:
                if hasattr(mdata, d):
                    mdata.__setattr__(d+'_', data.getncattr(d))
                else:
                    mdata.__setattr__(d, data.getncattr(d))
               
            # loadup all of the variables in the netCDF
            loadedVars={}
            for key in data.variables.keys():
                # load up metadata
                mdata[key] = {'long_name':data.variables[key].long_name,
                                'units':data.variables[key].units} #,
                                #'nc_dimensions':data.variables[key].dimensions}
                # from here group unique dimensions and act accordingly, 1D, 2D, 3D  
                if len(data.variables[key].dimensions) == 1:
                    # assuming basic time dimension
                    loadedVars[key] = data.variables[key][:] 
                if len(data.variables[key].dimensions) == 2:
                    # part of dataframe within dataframe
                    two_d_keys.append(key)
                    two_d_dims.append(data.variables[key].dimensions)
                    
            # we now have a list of keys that need to go into a dataframe,
            # could be more than one, collect unique dimensions for 2D keys
            for dim in set(two_d_dims):
                # get the name of the final data column
                # dimension naming follows name_dim_number, 
                # pull out name by finding last _ and tracking back
                obj_key_name = dim[1][ : -string.find(dim[1][::-1], '_')-5]
                # collect variable names associated with object
                obj_var_keys = []
                for tkey, tdim in zip(two_d_keys, two_d_dims):
                    if tdim == dim:
                        obj_var_keys.append(tkey)
                        
                # loop over first object dimension
                # preallocate dataframes to hold objects because it is faster
                init_frame = DataFrame(None)
                loop_list = [init_frame]*data.variables[obj_var_keys[0]].shape[0]
                for i, loop_frame in enumerate(loop_list):
                    #loop_frame = saved_frame.copy()
                    for key in obj_var_keys:
                        loop_frame[key] = data.variables[key][i,:]
                        
                    # if the object index uses unix time, process into datetime index    
                    if data.variables[obj_key_name+'_sample_index'].long_name == 'UNIX time':
                        # nansecond reolution from datetime64 can't be stored in netcdf3
                        # no 64-bit integers
                        # it is stored as a float, need to undo processing 
                        # due to precision loss, resolution limited to the microsecond
                        loop_frame.index = pds.to_datetime((1E6*loop_frame[obj_key_name+'_sample_index']).astype(int)*1000)
                    else:
                        loop_frame.index = loop_frame[obj_key_name+'_sample_index']
                    del loop_frame[obj_key_name+'_sample_index']  
                    #loop_list.append(loop_frame)                      
                # add object data to loaded data dictionary
                loadedVars[obj_key_name] = loop_list
                del loop_list
                
            # prepare dataframe index for this netcdf file    
            if index_label is not None:
                if unix_time:
                    loadedVars['_index'] = pds.to_datetime((loadedVars.pop(index_label)*1E6).astype(int)*1000)
                else:
                    loadedVars['_index'] = loadedVars.pop(index_label)
                running_store.append(loadedVars)
                running_idx += len(loadedVars['_index'])
            else:
                # keep a running integer index if none provided                          
                num = len(loadedVars[loadedVars.keys()[0]]) 
                # this only guaranteed to work if all variables share the same
                # first dimension  
                loadedVars['_index'] = np.arange(num) + running_idx                           
                running_store.append(loadedVars)     
                running_idx += num   

            if strict_meta:
                if saved_mdata is None:
                    saved_mdata = copy.deepcopy(mdata)
                elif (mdata != saved_mdata):
                    raise ValueError('Metadata across filenames is not the same.')
                    
    # combine all of the data loaded across files together
    # currently doesn't work if list of dicts of lists is provided
    # in other words, only one file at a time
    out = DataFrame.from_records(running_store[0], index='_index')
         
    return out, mdata        


def getyrdoy(date):
    """Return a tuple of year, day of year for a supplied datetime object."""
    #if date is not None:
    try:
        doy = date.toordinal()-datetime(date.year,1,1).toordinal()+1
    except AttributeError:
        raise AttributeError("Must supply a pandas datetime object or equivalent")
    else:
        return (date.year, doy)


def season_date_range(start, stop, freq='D'):
    """
    Return array of datetime objects using input frequency from start to stop
    
    Supports single datetime object or list, tuple, ndarray of start and 
    stop dates.
    
    freq codes correspond to pandas codes, D daily, M monthly, S secondly
    """
    
    if hasattr(start, '__iter__'):  
        #missing check for datetime
        season = pds.date_range(start[0], stop[0], freq=freq)
        for (sta,stp) in zip(start[1:], stop[1:]):
            season = season.append(pds.date_range(sta, stp, freq=freq))
    else:
        season = pds.date_range(start, stop, freq=freq)
    return season

#determine the median in 1 dimension
def median1D(self, bin_params, bin_label,data_label):

    bins = np.arange(bin_params[0],bin_params[1]+bin_params[2],bin_params[2])
    ans = 0.*bins[0:-1]
    ind = np.digitize(self.data[bin_label], bins)

    for i in xrange(bins.size-1):
        index, = np.where(ind==(i+1))
        if len(index)>0:
            ans[i] = self.data.ix[index, data_label].median()

    return ans


def create_datetime_index(year=None, month=None, doy=None, uts=None):
    """
    Create a timeseries index using supplied year,month,day, and ut in seconds.
    Leap seconds have no meaning here.

    Keywords:
        year - array/Series for year 
        month - array/Series for month
        doy - array/Series for day or day of year (use month=None)
        uts - array/Series of UTC seconds.
    Output:
        Pandas timeseries index.
    """
    # need a timeseries index for storing satellite data in pandas but
    # creating a datetime object for everything is too slow
    # so I calculate the number of nanoseconds elapsed since first sample, 
    # and create timeseries index from that. 
    # Factor of 20 improvement compared to previous method,
    # which itself was an order of magnitude faster than datetime.
 
    #get list of unique year, and month
    if not hasattr(year, '__iter__'):
        raise ValueError('Must provide an iterable for all inputs.')
    if len(year) == 0:
        raise ValueError('Length of array must be larger than 0.')
    year = year.astype(int)
    if month is None:
        month = np.ones(len(year), dtype=int)
    else:
        month = month.astype(int)
        
    if uts is None:
        uts = np.zeros(len(year))
    doy = doy.astype(int)
    # track changes in seconds
    uts_del = uts.copy()
    # determine where there are changes in year and month that need to be 
    # accounted for    
    _,idx = np.unique(year*100.+month, return_index=True)
    # create another index array for faster algorithm below
    idx2 = np.hstack((idx,len(year)+1))   
    # computes UTC seconds offset for each unique set of year and month
    for _idx,_idx2 in zip(idx[1:],idx2[2:]):
        temp = (datetime(year[_idx],month[_idx],1) - datetime(year[0],month[0],1))
        uts_del[_idx:_idx2] += temp.total_seconds()

    # add in UTC seconds for days, ignores existence of leap seconds
    uts_del += (doy-1)*86400
    # add in seconds since unix epoch to first day
    uts_del += (datetime(year[0],month[0],1)-datetime(1970,1,1) ).total_seconds()
    # going to use routine that defaults to nanseconds for epoch
    uts_del *= 1E9
    return pds.to_datetime(uts_del)
    
                                                
#def create_datetime_index_slow(year=None, doy=None, uts=None):
#    """
#    Create a Date Index for storing satellite date in pandas array.
#
#    scientific data will be loaded into a subclass of numpy called pandas
#    ,designed for time series.
#
#    Keywords:
#        yrdoy - array of yr*1000 + (Day of Year)
#        uts - array of ut seconds.
#    Output:
#        Pandas timearray containing satellite data.
#    """
#    #Need to create a time label for each of the measurements in time
#    #time labels use the python pandas.datetime functionality
#    #designed to work on a single day
#
#    #create datetime object with start date
#    #start_date = pds.datetime(yrs[0],1,1)+timedelta(days=(doys[0]-1),seconds=uts[0])
#    #create array of seconds since start time
#    
#    yr_offset = year - year[0]
#    if (year[0] % 4 == 0) & (year[0] % 100 != 0):
#        #leap year
#        uts_del = uts + (yr_offset*366. + doy)*86400.
#    else:
#        #non-leap year
#        uts_del = uts + (yr_offset*365. + doy)*86400.
#
#    uts_del -= uts_del[0]
#    #uts_del = array([(pds.datetime(year,1,1)+timedelta(days=(days-1),seconds=secs) - start_date).total_seconds() for (year,days,secs) in zip(yrs,doys,uts)])
#    #create an array of datetime seconds, then multiply by change in uts
#
#    #turns out this is super slow
##        timeIndex = pds.datetime(yrs[0],1,1) + pds.DateOffset(days=doys[0]-1, microseconds=uts[0]*1000000)
##        temp = pds.DateOffset(seconds=1)
##        timeIndex += uts_del*temp
##        return timeIndex
#
#    #create pandas.datetime object for first sample
#    #doesn't like inputs less than 1 second
#    start = pds.datetime(year[0],1,1)+pds.DateOffset(days=doy[0]-1,microseconds=uts[0]*1000000)
#    end = pds.datetime(year[-1],1,1)+pds.DateOffset(days=doy[-1]-1,microseconds=uts[-1]*1000000)
#
#    #create all times in between at a constant cadence
#    #creating pandas.datetime object for each time individually too slow
#    #thus I create a constant samplrate array
#    #and I find the satellite times closest to constant samplerate times
#    freq=1
#    timeIndex = pds.date_range(start, end, freq='%iL'%freq)
#
#    #create an index array that connects each satellite measurement time
#    #to a pandas.datetime object
#
#    #number of samples at freq samplerate to get to uts_del seconds
#    uts_del /= (freq/1000.)
#    #pick the closest integer
#    uts_del = np.floor(uts_del).astype('int64')
#    #this is now the index into the pandas.datetime array
#    #return the pandas.datetime array nearest each measurement to the
#    #closest 1 milisecond. Fails if instrument samplerate is close
#    #to this
#    return timeIndex[uts_del]

#def load_netcdf3_simple(fnames=None, strict_meta=False, **kwargs):
#    
#    import netCDF4
#
#    if fnames is None:
#        raise ValueError("Must supply a list of filenames")
#    if not hasattr(fnames, '__iter__'):
#        fnames = [fnames]    
#            
#    saved_mdata = None
#    idx = 0
#    temp_store=[]
#    for file in fnames:
#        with netCDF4.Dataset(file, mode='r', format='NETCDF3_64BIT') as data:
#            # build up dictionary will all ncattrs
#            # and add those attributes to a pysat meta object
#            ncattrsList = data.ncattrs()
#            mdata = meta.Meta()
#            for d in ncattrsList:
#                if hasattr(mdata, d):
#                    mdata.__setattr__(d+'_', data.getncattr(d))
#                else:
#                    mdata.__setattr__(d, data.getncattr(d))
#               
#            #loadup all of the variables in the netCDF
#            loadedVars={}
#            keys = data.variables.keys()
#            for key in keys:
#                # load up metadata
#                mdata[key] = {'long_name':data.variables[key].long_name,
#                                'units':data.variables[key].units,
#                                'dimensions':data.variables[key].dimensions}
#                # from here group unique dimensions and act accordingly, 1D, 2D, 3D 
#                # cheating at the moment  
#                loadedVars[key] = data.variables[key][:] 
#            num = len(loadedVars[keys()[0]])                               
#            temp_store.append(pds.DataFrame.from_dict(loadedVars,
#                                index=np.arange(num) + idx ))     
#            idx += num   
#
#            if strict_meta:
#                if saved_mdata is None:
#                    saved_mdata = copy.deepcopy(mdata)
#                elif (mdata != saved_mdata):
#                    raise ValueError('Metadata across filenames is not the same.')
#                    
#    return pds.DataFrame(temp_store), mdata        
        #out_data.close()



import string
import os
import weakref
import re
import glob
import numpy as np
import pandas as pds
from pysat import data_dir as data_dir

class Files(object):
        
    def __init__(self, sat):

        self._sat = weakref.proxy(sat)
        self.base_path = os.path.join(os.getenv('HOME'), '.pysat')
        self.start_date = None
        self.stop_date = None
        self.files = pds.Series(None)
        self.data_path = os.path.join(data_dir,self._sat.platform, self._sat.name, self._sat.tag)
        
        if self._sat.platform != '':
            info = self._load()#self._sat.platform, self._sat.name, self._sat.tag)
       	    if info is not False:
       	        self._attach_files(info)
       	    else:
       	        print "pysat is searching for the requested instrument's files."
   	        # couldn't find stored info, load file list and then store
   	        info = self._sat._list_rtn(tag=self._sat.tag, data_path=self.data_path)
                if info is not None:
                    if len(info) > 0:
                        info = self._remove_data_dir_path(info)	
       	                self._attach_files(info)
       	                self._store()

    def _attach_files(self, files_info):
        """Attaches info returned by instrument list_files routine to Instrument object."""

	if (len(files_info.unique()) != len(files_info)): 
	    raise ValueError('List of files must have unique datetimes.')
   
	self.files = files_info.sort_index()
	date = files_info.index[0] 
	self.start_date = pds.datetime(date.year, date.month, date.day)
	date = files_info.index[-1] 
	self.stop_date = pds.datetime(date.year, date.month, date.day)

                
    def _store(self, dir=None):
        """Store currently loaded filelist for instrument onto filesystem"""
        name = ''.join((self._sat.platform,'_',self._sat.name,'_',self._sat.tag,
                        '_stored_file_info.txt'))
        if dir is None:
            name = os.path.join(self.base_path, name)
        else:
            name = os.path.join(dir, name)
        try:
            self.files.to_csv(name)
            return True
        except IOError:
            return False	


    def _load(self):
        """Load stored filelist and return as Pandas Series"""
        fname = ''.join((self._sat.platform,'_',self._sat.name,'_',
                        self._sat.tag, '_stored_file_info.txt'))
        fname = os.path.join(self.base_path, fname)
        try:
            data = pds.Series.from_csv(fname, index_col=0)
        except IOError:
            return False
        else:
            return data


    def refresh(self, store=False):
        """Refresh loaded instrument filelist by searching filesystem."""
        info = self._sat._list_rtn(tag=self._sat.tag, data_path=self.data_path)
        info = self._remove_data_dir_path(info)
        self._attach_files(info)
        if store:
            self._store()
        
    
    def get_new(self):
        """List all files on filesystem that weren't there the last time
           the file list was stored.
        """
        storedInfo = self._load()
        if storedInfo is not False:
            newInfo = self._sat._list_rtn(tag = self._sat.tag, data_path=self.data_path)
            newInfo = self._remove_data_dir_path(newInfo)
            boolArr = newInfo.isin(storedInfo) 
            newFiles = newInfo[~boolArr]
            return newFiles
        else:
            print 'No previously stored files that we may compare to.'
            return False

    def get_index(self, fname):
        """Return index in objects file_list for a particular filename. """
        idx, = np.where(fname == self.files)
        if len(idx) == 0:
            #filename not in index, try reloading files from disk
            self.refresh()
            idx, = np.where(fname == self.files)
            if len(idx) == 0:
		raise IOError('Could not find supplied file on disk')
	return idx 
	
    #convert this to a normal get so files[in:in2] gives the same as requested here
    #support slicing via date, filename, and index
    #filename is inclusive slicing, date and index are normal non-inclusive end point
    #is this hidden behavior good though?????	
    # I havent actually implemented difference in slicing yet   	  	
    def __getitem__(self, key):
        if isinstance(key, slice):
            try:
                out = self.files.ix[key]
	    except IndexError:
	        raise IndexError('Date requested outside file bounds.')                
            if len(out) > 1:
                if out.index[-1] >= key.stop:
                    return out[:-1]
                else:
                    return out
            elif len(out) == 1:
                if out.index[0] >= key.stop:
                    return pds.Series([])
                else:
                    return out
            else:
                return out
        else:
            raise ValueError('Not implemented yet.')         
        #if isinstance(key, tuple):
        #    if len(key) == 2:
        #        start = key[0]
        #        end = key[1]
        #    else:
        #        raise ValueError('Must input 2 and only 2 items/iterables')
                
    def get_file_array(self, start, end):
        """Return a list of filenames between and including start and end.
        
        paramters:
            start: list or single string of filenames for start of list
            stop: list or single string of filenames inclusive end of list
        returns:
            list of filenames between and including start and end over all
            intervals. 
            """
        if hasattr(start, '__iter__') & hasattr(end, '__iter__'):
            files = []
            for (sta,stp) in zip(start, end):
                id1 = self.get_index(sta)
		id2 = self.get_index(stp)
		files.extend(self.files[id1 : id2+1])
	elif hasattr(start, '__iter__') | hasattr(end, '__iter__'):
	    raise ValueError('Either both or none of the inputs need to be iterable')
        else:
            id1 = self.get_index(start)
	    id2 = self.get_index(end)
	    files = self.files[id1:id2+1].to_list()   
	return files
	                      
    def _remove_data_dir_path(self, inp=None):
        """Remove the data directory path from filenames"""
        # need to add a check in here to make sure data_dir path is actually in
        # the filename
        if inp is not None:
            match = os.path.join(self.data_path,'')
            num = len(match)	 
            return inp.apply(lambda x: x[num:])	
        
    @classmethod    
    def from_os(cls, data_path=None, format_str=None, 
                two_digit_year_break=None):
        """
        Produces a list of files and and formats it for Files class.
        """
        from utils import create_datetime_index
        if (format_str is None):
            raise ValueError("Must supply a filename template (format_str).")
        if (data_path is None):
            raise ValueError("Must supply instrument directory path (dir_path)")
        
        # parse format string to figure out the search string to use
        # to identify files in the filesystem
        search_str = ''
        form = string.Formatter()
        keys = []
        snips = []
        length = []
        stored = {'year':[], 'month':[], 'day':[], 'hour':[], 'min':[], 'sec':[]}
        for snip in form.parse(format_str):
            search_str += snip[0]
            snips.append(snip[0])
            if snip[1] is not None:
                keys.append(snip[1])
                search_str += '*'
                # try and determine formatting width
                temp = re.findall(r'\d+', snip[2])
                if temp:
                    # there are items, try and grab width
                    for i in temp:
                        if i != 0:
                            length.append(int(i))
                            break
                else:
                    raise ValueError("Couldn't determine formatting width")

        abs_search_str = os.path.join(data_path, search_str)
        files = glob.glob(abs_search_str)   
        
        # we have a list of files, now we need to extract the date information        
        # code below works, but only if the size of file string 
        # remains unchanged
        
        # determine the loactaions the date information in a filename is stored
        # use these indices to slice out date from loaded filenames
        #test_str = format_str.format(**periods)  
        if len(files) > 0:  
            idx = 0
            begin_key = []
            end_key = []
            for i,snip in enumerate(snips):
                #if snip is not None:
                #begin_lit.append(test_str.find(snip,idx) )
                #end_lit.append(begin_lit[-1] + len(snip))
                #idx = end_lit[-1]
                idx += len(snip)
                if i < (len(length)):
                    begin_key.append(idx)
                    idx += length[i]
                    end_key.append(idx)
            max_len = idx
            # setting up negative indexing to pick out filenames
            key_str_idx = [np.array(begin_key, dtype=int) - max_len, 
                            np.array(end_key, dtype=int) - max_len]
            #print key_str_idx
            # need to parse out dates for datetime index
            for i,temp in enumerate(files):
                for j,key in enumerate(keys):
                    val = temp[key_str_idx[0][j]:key_str_idx[1][j]]
                    stored[key].append(val)
    
            # convert to numpy arrays
            for key in stored.keys():
                stored[key] = np.array(stored[key]).astype(int)
                if len(stored[key]) == 0:
                    stored[key]=None
            
            # deal with the possibility of two digit years
            # years above or equal to break are considered to be 1900+
            # years below break are considered to be 2000+
            if two_digit_year_break is not None:
                idx, = np.where(stored['year'] >= two_digit_year_break)
                stored['year'][idx] = stored['year'][idx] + 1900
                idx, = np.where(stored['year'] < two_digit_year_break)
                stored['year'][idx] = stored['year'][idx] + 2000            
    
            index = create_datetime_index(year=stored['year'], month=stored['month'], 
                                    doy=stored['day'], uts=stored['sec'])

            return pds.Series(files, index=index)
        else:
            print ("Unable to find any files. If you have the necessary files "+
                    "please check pysat settings and file locations.") 

        


def scatterplot(inst, start, stop, labelx, labely, data_label, datalim, xlim=None, ylim=None):
    """Return a scatterplot over a season.

    Arguments:
        Start: Start Date(s) - datetime object
        Stop: Stop Date(s) - datetime object
        labelx: string identifying data product for x-axis
        labely: string identifying data product for y-axis
        data_label: string identifying data product to be scatter plotted
        datalim: np array with plot limits for data_label

    Summary:
        Returns a scatter plot of data_label as a function of label1 and label2 over the season delineated by start and stop datetime objects."""

    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib import cm
    import matplotlib.colors as colors

    #interactive plotting off
    plt.ioff()
    
    # do loop to iterate over given days
    # start data in start, end day in stop
    season = pysat.getDateArray(start, stop, freq='D')
    
    #create figures for plotting
    figs = []
    axs = []
    
    #check if doing multiple data quantities
    if type(data_label) is str:
        #max of one data type
        figs.append( plt.figure() )
        ax1 = figs[0].add_subplot(211, projection='3d')
        ax2 = figs[0].add_subplot(212)
        axs.append((ax1, ax2))
        plt.suptitle(data_label)
        if xlim is not None:
            ax1.set_xlim(xlim)
	    ax2.set_xlim(xlim)	
        if ylim is not None:
            ax1.set_ylim(ylim)
	    ax2.set_ylim(ylim)	
        data_label = [data_label]
        plt.hold(True)
    else:
        #multiple data to be plotted
        for i in np.arange(len(data_label)):
            figs.append( plt.figure() )
            ax1 = figs[i].add_subplot(211, projection='3d')
            ax2 = figs[i].add_subplot(212)
            axs.append((ax1, ax2))
            #axs.append( figs[i].add_subplot(111, projection='3d') )
            plt.suptitle(data_label[i])              
            if xlim is not None:
               ax1.set_xlim(xlim)
	       ax2.set_xlim(xlim)
            if ylim is not None:
               ax1.set_ylim(ylim)
	       ax2.set_ylim(ylim)
            plt.hold(True)
                
                
    #norm method so that data may be scaled to colors appropriately
    norm = colors.Normalize(vmin=datalim[0], vmax=datalim[1])    
    p=[i for i in np.arange(len(figs))]    
    for i,date in enumerate(season):
        #load data
        inst.getData(date=date)
        for j, (fig, ax) in enumerate(zip(figs, axs)):
            if len(inst.data) > 0:
                if ((len(inst.data[labelx]) > 0) & (len(inst.data[labely]) > 0) & (len(inst.data[data_label[j]]) > 0)):
                    p[j]=ax[0].scatter(inst.data[labelx], inst.data[labely], inst.data[data_label[j]], zdir='z', c=inst.data[data_label[j]], cmap=cm.jet, norm=norm, linewidth='0', edgecolors=None)
                    ax[1].scatter(inst.data[labelx], inst.data[labely], c=inst.data[data_label[j]], cmap=cm.jet, norm=norm, linewidth=0.00000000001, alpha=0.5, edgecolor=None)
                
    for j, (fig, ax) in enumerate(zip(figs, axs)):
	try:
            cbar = plt.colorbar(p[j],ax=ax[0], label='Amplitude (m/s)')
        except:
            print 'Tried colorbar but failed, thus no colorbar.'
        ax[0].elev=30.
    #interactive plotting back on    
    plt.ion()
    
    return figs


"""
Occurrence probability routines, daily or by orbit.
"""

import numpy as np
import matplotlib.pyplot as plt

def daily2D(inst, bin1, label1, bin2, label2, data_label, gate, returnBins=False):
    """
    Daily Occurrence Probability of data_label over a season binned by two
    parameters. If data_label is greater than gate atleast once per day, 
    then a 100% occurrence probability results. Season delineated by the bounds
    attached to Instrument object.

    Arguments:
        binx: [min, max, number of bins]
        labelx: string identifying data product for binx
        data_label: list of strings identifying data product to calculate 
                    occurrence probability
        gate: list of values that data_label must achieve to be counted as an occurrence

    Summary:
         Returns a dict with numpy arrays, 'prob' for the probability and
         'count' for the number of days with any data.'binx' and 'biny' are also
         returned if requested. Note that arrays are organized for direct 
         plotting, y values along rows, x along columns
    """
    
    return _occurrence2D(inst, bin1, label1, bin2, label2, data_label, gate,
                        by_orbit=False, returnBins=returnBins)


def by_orbit2D(inst, bin1, label1, bin2, label2, data_label, gate, returnBins=False):
    """
    Occurrence Probability of data_label orbit-by-orbit over a season binned by two
    parameters. If data_label is greater than gate atleast once per day, 
    then a 100% occurrence probability results. Season delineated by the bounds
    attached to Instrument object.

    Arguments:
        binx: [min, max, number of bins]
        labelx: string identifying data product for binx
        data_label: list of strings identifying data product to calculate 
                    occurrence probability
        gate: list of values that data_label must achieve to be counted as an occurrence

    Summary:
         Returns a dict with numpy arrays, 'prob' for the probability and
         'count' for the number of days with any data.'binx' and 'biny' are also
         returned if requested. Note that arrays are organized for direct 
         plotting, y values along rows, x along columns
    """
    
    return _occurrence2D(inst, bin1, label1, bin2, label2, data_label, gate,
                        by_orbit=True, returnBins=returnBins)

                                                                                                
def _occurrence2D(inst, bin1, label1, bin2, label2, data_label, gate, 
                by_orbit=False, returnBins=False):
    if not hasattr(data_label, '__iter__'):
        raise ValueError('Data label must be list-like group of variable names.')
    if not hasattr(gate, '__iter__'):
        raise ValueError('Gate levels must be list-like group of variable names.')
    if len(gate) != len(data_label):
        raise ValueError('Must have a gate value for each data_label')

    #create bins
    binx = np.linspace(bin1[0], bin1[1], bin1[2]+1)
    biny = np.linspace(bin2[0], bin2[1], bin2[2]+1)

    numx = len(binx)-1
    numy = len(biny)-1
    numz = len(data_label)

    # create arrays to store all values
    total = np.zeros((numz, numy, numx))
    hits = np.zeros((numz, numy, numx))
    if by_orbit:
        inst.load(date=inst.bounds[0][0])
        iterator = inst.orbits
    else:
        iterator = inst
    # do loop to iterate over given iterator
    for i,inst in enumerate(iterator):
        if len(inst.data) != 0:
            xind = np.digitize(inst.data[label1], binx)-1
            for xi in xrange(numx):
                xindex, = np.where(xind==xi)
                if len(xindex)>0:
                    yData = inst.data.ix[xindex]
                    yind = np.digitize(yData[label2], biny)-1
                    for yj in xrange(numy):
                        yindex, = np.where(yind==yj)
                        if len(yindex) > 0:
                            # iterate over the different data_labels
                            for zk in xrange(numz):
                                zdata = yData.ix[yindex,data_label[zk]]
                                idx, = np.where(np.isfinite(zdata))
                                if len(idx) > 0:
                                    total[zk,yj,xi] += 1.
                                    idx, = np.where( zdata > gate[zk]  )
                                    if len(idx) > 0:
                                        hits[zk,yj,xi] += 1.
                               
    # all of the loading and storing data is done
    # get probability
    prob = hits/total
    # make nicer dictionary output
    output = {}
    for i,label in enumerate(data_label):
        output[label] = {'prob': prob[i,:,:], 'count':total[i,:,:]}
        if returnBins:
            output[label]['binx'] = binx
            output[label]['biny'] = biny
    # clean up
    del iterator
    return output
        
        
def daily3D(inst, bin1, label1, bin2, label2, bin3, label3, 
            data_label, gate, returnBins=False):
    """
    Daily Occurrence Probability of data_label over a season binned by two
    parameters. If data_label is greater than gate atleast once per day, 
    then a 100% occurrence probability results. Season delineated by the bounds
    attached to Instrument object.

    Arguments:
        binx: [min, max, number of bins]
        labelx: string identifying data product for binx
        data_label: list of strings identifying data product to calculate 
                    occurrence probability
        gate: list of values that data_label must achieve to be counted as an occurrence

    Summary:
         Returns a dict with numpy arrays, 'prob' for the probability and
         'count' for the number of days with any data.'binx','biny', and 'binz'
         are also returned if requested. Note that arrays are organized for direct 
         plotting, z,y,x. 
    """
        
    return _occurrence3D(inst, bin1, label1, bin2, label2, bin3, label3, 
                        data_label, gate, returnBins=returnBins, by_orbit=False)
           

def by_orbit3D(inst, bin1, label1, bin2, label2, bin3, label3, 
                data_label, gate, returnBins=False):
    """
    Occurrence Probability of data_label over a season orbit-by-orbit binned by two
    parameters. If data_label is greater than gate atleast once per day, 
    then a 100% occurrence probability results. Season delineated by the bounds
    attached to Instrument object.

    Arguments:
        binx: [min, max, number of bins]
        labelx: string identifying data product for binx
        data_label: list of strings identifying data product to calculate 
                    occurrence probability
        gate: list of values that data_label must achieve to be counted as an occurrence

    Summary:
         Returns a dict with numpy arrays, 'prob' for the probability and
         'count' for the number of days with any data.'binx','biny', and 'binz'
         are also returned if requested. Note that arrays are organized for direct 
         plotting, z,y,x. 
    """
        
    return _occurrence3D(inst, bin1, label1, bin2, label2, bin3, label3, 
                        data_label, gate, returnBins=returnBins, by_orbit=True)

                                     
                                                                                                               
def _occurrence3D(inst, start, stop, bin1, label1, bin2, label2, bin3, label3, 
                    data_label, gate, returnBins=False, by_orbit=False):

    if not hasattr(data_label, '__iter__'):
        raise ValueError('Data label must be list-like group of variable names.')
    if not hasattr(gate, '__iter__'):
        raise ValueError('Gate levels must be list-like group of variable names.')
    if len(gate) != len(data_label):
        raise ValueError('Must have a gate value for each data_label')

    #create bins
    binx = np.linspace(bin1[0], bin1[1], bin1[2]+1)
    biny = np.linspace(bin2[0], bin2[1], bin2[2]+1)
    binz = np.linspace(bin3[0], bin3[1], bin3[2]+1)

    numx = len(binx)-1
    numy = len(biny)-1
    numz = len(binz)-1
    numd = len(data_label)

    #create array to store all values before taking median
    yarr = np.arange(numy)
    xarr = np.arange(numx)
    zarr = np.arange(numz)
    darr = np.arange(numd)

    total = np.zeros((numd, numz, numy, numx))
    hits = np.zeros((numd, numz, numy, numx))

    if by_orbit:
        iterator = inst.orbits
    else:
        iterator = inst
    # do loop to iterate over given season
    for i,sat in enumerate(iterator):

        if len(sat.data) != 0:
            xind = np.digitize(sat.data[label1], binx)-1
            for xi in xarr:
                xindex, = np.where(xind==xi)
                if len(xindex)>0:
                    yData = sat.data.ix[xindex]
                    yind = np.digitize(yData[label2], biny)-1
                    for yj in yarr:
                        yindex, = np.where(yind==yj)
                        if len(yindex) > 0:
                            zData = yData.ix[yindex]
                            zind = np.digitize(zData[label3], binz)-1
                            for zk in zarr:
                                zindex, = np.where(zind==zk)
                                if len(zindex) > 0:
                                    for di in darr:
                                        ddata = zData.ix[zindex,data_label[di]]
                                        idx, = np.where(np.isfinite(ddata))
                                        if len(idx) > 0:
                                            total[di,zk,yj,xi] += 1
                                            idx, = np.where( ddata > gate[di]  )
                                            if len(idx) > 0:
                                                hits[di,zk,yj,xi] += 1
                                
                               
    #all of the loading and storing data is done
    #prob = np.zeros((numz, numy, numx))
    prob = hits/total

    # make nicer dictionary output
    output = {}
    for i,label in enumerate(data_label):
        output[label] = {'prob': prob[i,:,:,:], 'count':total[i,:,:,:]}
        if returnBins:
            output[label]['binx'] = binx
            output[label]['biny'] = biny
            output[label]['binz'] = biny
    # clean up
    del iterator
    return output    


# -*- coding: utf-8 -*-
"""
Created on Thu Jul 19 17:03:47 2012

@author: musicalphysics
"""
import pandas as pds
import numpy as np
import pysat

platform = 'pysat'
name = 'testing'

def init(self):
    self.new_thing=True
    
def load(fnames, tag=None):
    # create an artifical satellite data set
    parts = fnames[0].split('/')
    yr = int('20'+parts[-1][0:2])
    month = int(parts[-3])
    day = int(parts[-2])
    date = pysat.datetime(yr,month,day)
    num = 86400 #int(tag)
    uts = np.arange(num)
    data = pysat.DataFrame(uts, columns=['uts'])


    # need to create simple orbits here. Have start of first orbit 
    # at 2009,1, 0 UT. 14.84 orbits per day	
    time_delta = date  - pysat.datetime(2009,1,1) 
    uts_root = np.mod(time_delta.total_seconds(), 5820)
    mlt = np.mod(uts_root+np.arange(num), 5820)*(24./5820.)
    data['mlt'] = mlt
    
    # do slt, 20 second offset from mlt
    uts_root = np.mod(time_delta.total_seconds()+20, 5820)
    data['slt'] = np.mod(uts_root+np.arange(num), 5820)*(24./5820.)

    index = pds.date_range(date,date+pds.DateOffset(hours=23,minutes=59,seconds=59),freq='S')
    data.index=index
    return data, pysat.Meta()

def list_files(tag=None, data_path=None):
    """Produce a fake list of files spanning a year"""
    
    index = pds.date_range(pysat.datetime(2008,1,1), pysat.datetime(2010,12,31)) 
    names = [ data_path+'/'+date.strftime('%D')+'.nofile' for date in index]
    return pysat.Series(names, index=index)
    
def download(start, stop, data_path=None,user=None, password=None):
    pass
#def meta():
#    
#    code = ['uts', 'yrdoy', 'mlt', 'slt']
#    index = np.arange(len(code))
#    label = ['uts', 'Year Day of Year', 'Magnetic Local Time', 'Solar Local Time']
#    units = ['s', ' ', 'hours', 'hours']
#    description = ['', '', '', '']
#    
#    d = {'code':code, 'index':index, 'label':label, 'units':units, 'description':description}
#    
#    return pysat.DataFrame(d, index=d['code'])


# -*- coding: utf-8 -*-
#Test some of the basic _core functions

import pysat
import pandas as pds
from nose.tools import assert_raises, raises
import nose.tools
import pysat.instruments.pysat_testing

class TestBasics:
    def setup(self):
        reload(pysat.instruments.pysat_testing)
	'''Runs before every method to create a clean testing setup.'''
        self.testInst = pysat.Instrument('pysat', 'testing', '10', 'clean')

    def teardown(self):
        '''Runs after every method to clean up previous testing.'''
        del self.testInst
    
    def test_basic_instrument_load(self):
        '''Test if the correct day is being loaded (checking object date).'''
        self.testInst.load(2009,1)
        assert self.testInst.date == pds.datetime(2009,1,1)	

    def test_next_load_default(self):
        '''Test if first day is loaded by default when first invoking .next.'''
        self.testInst.next()
        assert self.testInst.date == pds.datetime(2008,1,1)

    def test_prev_load_default(self):
        '''Test if last day is loaded by default when first invoking .prev.'''
        self.testInst.prev()
        assert self.testInst.date == pds.datetime(2010,12,31)
    
    def test_filename_load(self):
        '''Test if file is loadable by filename, relative to top_data_dir/platform/name/tag'''
        self.testInst.load(fname='12/31/10.nofile')
        assert self.testInst.data.index[0] == pds.datetime(2010,12,31)
        
    def test_instrument_init(self):
        """Test if init function supplied by instrument can modify object"""
        assert self.testInst.new_thing==True

    @raises(StopIteration)
    def test_left_bounds_with_prev(self):
        '''Test if passing bounds raises StopIteration.'''
        self.testInst.next()
        self.testInst.prev()
        self.testInst.prev()        
        
    @raises(StopIteration)
    def test_right_bounds_with_next(self):
        '''Test if passing bounds raises StopIteration.'''
        self.testInst.prev()
        self.testInst.next()
        self.testInst.next()        

    def test_basic_instrument_load_data(self):
        '''Test if the correct day is being loaded (checking data).'''
        self.testInst.load(2009,1)
        assert self.testInst.data.index[0] == pds.datetime(2009,1,1,0,0,0)

    def test_basic_instrument_load_leap_year(self):
        '''Test if the correct day is being loaded (Leap-Year).'''
        self.testInst.load(2008,366)
        assert self.testInst.date == pds.datetime(2008,12,31)	

    def test_getyrdoy_1(self):
	'''Test the date to year, day of year code functionality'''
        date = pds.datetime(2009,1,1)
        yr, doy = pysat.utils.getyrdoy(date)
        assert ((yr == 2009) & (doy == 1))

    def test_getyrdoy_leap_year(self):
	'''Test the date to year, day of year code functionality (leap_year)'''
        date = pds.datetime(2008,12,31)
        yr, doy = pysat.utils.getyrdoy(date)
        assert ((yr == 2008) & (doy == 366)) 


    def test_custom_instrument_load(self):
        '''
        Test if the correct day is being loaded (End-to-End), 
        with no instrument file but routines are passed.
        '''
        import pysat.instruments.pysat_testing as test
        testInst = pysat.Instrument(inst_module=test, tag='1000', clean_level='clean')
        testInst.load(2009,32)
        assert testInst.date == pds.datetime(2009,2,1)	
        
    @raises(AttributeError)
    def test_custom_instrument_load_2(self):
        '''
        Test if an exception is thrown correctly if there is no 
        instrument file and supplied routines are incomplete.
        '''
        import pysat.instruments.pysat_testing as test
        del test.list_files
        testIn = pysat.Instrument(inst_module=test, tag='1000', clean_level='clean')
        testIn.load(2009,1)

    @raises(AttributeError)
    def test_custom_instrument_load_3(self):
        '''
        Test if an exception is thrown correctly if there is no 
        instrument file and supplied routines are incomplete.
        '''
        import pysat.instruments.pysat_testing as test
        del test.load
        testIn = pysat.Instrument(inst_module=test, tag='1000', clean_level='clean')
        testIn.load(2009,1)

    def test_data_padding(self):
        te = pysat.Instrument('pysat','testing', '10000', pad=True, minutes=5)
        te.load(2009,1, verifyPad=True)
        print te.data.index[0]
        print te.data.index[-1]
        assert ( (te.data.index[0] == te.date - pds.DateOffset(minutes=5)) & 
                (te.data.index[-1] == te.date + pds.DateOffset(hours=23,minutes=59,seconds=59) + pds.DateOffset(minutes=5)) )
            
    def test_data_padding_removal(self):
        te = pysat.Instrument('pysat','testing', '10000', pad=True, minutes=5)
        te.load(2009,1)
        assert (te.data.index[0] == te.date ) & (te.data.index[-1] == te.date + pds.DateOffset(hour=23, minutes=59,seconds=59) )


import pysat
import pandas as pds
from nose.tools import assert_raises, raises
import nose.tools

from dateutil.relativedelta import relativedelta as relativedelta
class TestOrbits:
    
    def setup(self):
	'''Runs before every method to create a clean testing setup.'''
        self.testInst = pysat.Instrument('pysat','testing', '86400', 'clean', orbit_index='mlt')

    def teardown(self):
        '''Runs after every method to clean up previous testing.'''
        del self.testInst 
    
    def test_single_orbit_call_orbit_starts_0_UT_using_next(self):
        self.testInst.load(2009,1)
	self.testInst.orbits.next()
	ans = (self.testInst.data.index[0] == pds.datetime(2009,1,1))
	ans2 = (self.testInst.data.index[-1] == (pds.datetime(2009,1,1,1,36,59) ))
	print self.testInst.data.index[0]
        print self.testInst.data.index[-1]
	print ans
	print ans2
	assert ans & ans2

    def test_single_orbit_call_orbit_starts_0_UT_using_prev(self):
        
        self.testInst.load(2009,1)
	self.testInst.orbits.prev()
	ans = (self.testInst.data.index[0] == (pds.datetime(2009,1,1)-relativedelta(hours=1, minutes=37)) )
	ans2 = (self.testInst.data.index[-1] == (pds.datetime(2009,1,1)-relativedelta(seconds=1) ))
	assert ans & ans2 
	
    def test_single_orbit_call_orbit_starts_off_0_UT_using_next(self):
        from dateutil.relativedelta import relativedelta as relativedelta
        self.testInst.load(2008,366)
	self.testInst.orbits.next()
	print self.testInst.data.index[0], pds.datetime(2008,12,30, 23, 45), self.testInst.data.index[-1], (pds.datetime(2008,12,30, 23, 45)+relativedelta(hours=1, minutes=36, seconds=59) )
	ans = (self.testInst.data.index[0] == pds.datetime(2008,12,30, 23, 45))
	ans2 = (self.testInst.data.index[-1] == (pds.datetime(2008,12,30, 23, 45)+relativedelta(hours=1, minutes=36, seconds=59) ))
	assert ans & ans2

    def test_single_orbit_call_orbit_starts_off_0_UT_using_prev(self):
        self.testInst.load(2008,366)
	self.testInst.orbits.prev()
	ans = (self.testInst.data.index[0] == (pds.datetime(2008,12,30, 23, 45)-relativedelta(hours=1, minutes=37)) )
	ans2 = (self.testInst.data.index[-1] == (pds.datetime(2008,12,30, 23, 45)-relativedelta(seconds=1) ))
	assert ans & ans2 
	
    def test_repeated_orbit_calls_symmetric_single_day_0_UT(self):
        self.testInst.load(2009,1)
	self.testInst.orbits.next()
        control = self.testInst.copy()
	for j in xrange(10):
	    self.testInst.orbits.next()
	for j in xrange(10):
	    self.testInst.orbits.prev()
	assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_multi_day_0_UT(self):
        self.testInst.load(2009,1)
	self.testInst.orbits.next()
        control = self.testInst.copy()
	for j in xrange(20):
	    self.testInst.orbits.next()
	for j in xrange(20):
	    self.testInst.orbits.prev()
        print control.data.mlt
        print self.testInst.data.mlt
	assert all(control.data == self.testInst.data)      

    def test_repeated_orbit_calls_symmetric_single_day_off_0_UT(self):
        self.testInst.load(2008,366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
	for j in xrange(10):
	    self.testInst.orbits.next()
	for j in xrange(10):
	    self.testInst.orbits.prev()
	assert all(control.data == self.testInst.data)
	
    def test_repeated_orbit_calls_symmetric_multi_day_off_0_UT(self):
        self.testInst.load(2008,366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
	for j in xrange(20):
	    self.testInst.orbits.next()
	for j in xrange(20):
	    self.testInst.orbits.prev()
	assert all(control.data == self.testInst.data)        	

    def test_repeated_orbit_calls_antisymmetric_multi_day_off_0_UT(self):
        self.testInst.load(2008,366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
	for j in xrange(10):
	    self.testInst.orbits.next()
	for j in xrange(20):
	    self.testInst.orbits.prev()
	for j in xrange(10):
	    self.testInst.orbits.next()
	print control.data.mlt
        print self.testInst.data.mlt
	assert all(control.data == self.testInst.data) 

    def test_repeated_orbit_calls_antisymmetric_multi_multi_day_off_0_UT(self):
        self.testInst.load(2008,366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
	for j in xrange(20):
	    self.testInst.orbits.next()
	for j in xrange(40):
	    self.testInst.orbits.prev()
	for j in xrange(20):
	    self.testInst.orbits.next()
	print control.data.mlt
        print self.testInst.data.mlt
	assert all(control.data == self.testInst.data )	                    

    def test_repeated_orbit_calls_antisymmetric_multi_day_0_UT(self):
        self.testInst.load(2009,1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
	for j in xrange(10):
	    self.testInst.orbits.next()
	for j in xrange(20):
	    self.testInst.orbits.prev()
	for j in xrange(10):
	    self.testInst.orbits.next()
	print control.data.mlt
        print self.testInst.data.mlt
	assert all(control.data == self.testInst.data )

    def test_repeated_orbit_calls_antisymmetric_multi_multi_day_0_UT(self):
        self.testInst.load(2009,1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
	for j in xrange(20):
	    self.testInst.orbits.next()
	for j in xrange(40):
	    self.testInst.orbits.prev()
	for j in xrange(20):
	    self.testInst.orbits.next()
	print control.data.mlt
        print self.testInst.data.mlt
	assert all(control.data == self.testInst.data )


# -*- coding: utf-8 -*-
import os
# make sure a pysat directory exists
if not os.path.isdir(os.path.join(os.getenv('HOME'), '.pysat')):
    # create directory
    print ''.join(('Creating .pysat directory. Run pysat.utils.set_data_dir to set the path',
                   ' to top-level directory containing science data.'))
    os.mkdir(os.path.join(os.getenv('HOME'), '.pysat'))
    with open(os.path.join(os.getenv('HOME'), '.pysat', 'data_path.txt'),'w') as f:
        f.write('')
    data_dir=''
else:
    # load up stored data path
    with open(os.path.join(os.getenv('HOME'), '.pysat', 'data_path.txt'),'r') as f:
        data_dir = f.readline()
    if data_dir == '':
        print 'Run pysat.utils.set_data_dir to set the path to top-level directory containing science data.'
del f

from pandas import Panel, DataFrame, Series, datetime
import instruments
import utils
from _instrument import Instrument
import ssnl
from _meta import Meta
from _files import Files


__all__ = ['ssnl','instruments', 'utils','meta','files']





# -*- coding: utf-8 -*-
"""
Supports downloading, loading, and cleaning measurements from the
Vefctor Electric Field Instrument (VEFI) onboard the Communication and Navigation Outage Forecasting
System (C/NOFS) satellite.

current tags:
    'dc_b' : 1 second DC magnetometer data
"""

import pandas as pds
import numpy as np
import pysat
import os

import spacepy
from spacepy import pycdf
import pysat


def list_files(tag=None, data_path=None):
    """Return a Pandas Series of every file for chosen satellite data"""
    if data_path is not None:
        if tag == 'dc_b':
            return pysat.Files.from_os(data_path=data_path, 
            format_str='cnofs_vefi_bfield_1sec_{year:04d}{month:02d}{day:02d}_v05.cdf')
        else:
            raise ValueError('Unknown tag')
    else:
        raise ValueError ('A directory must be passed to the loading routine for VEFI')
            
# define metadata for VEFI instrument                              
meta = pysat.Meta()
meta['Epoch'] = {'long_name':'Epoch', 'units':'s'}
meta['year'] = {'long_name':'Year', 'units':'years'}
meta['dayofyear'] = {'long_name':'Day of Year', 'units':'days'}
meta['B_north'] = {'long_name':'Magnetic Field (North)', 'units':'nT'}
meta['B_up'] = {'long_name':'Magnetic Field (Up)', 'units':'nT'}
meta['B_west'] = {'long_name':'Magnetic Field (West)', 'units':'nT'}
meta['B_flag'] = {'long_name':'Flag', 'units':''}
meta['B_IGRF_north'] = {'long_name':'IGRF Magnetic Field (North)', 'units':'nT'}
meta['B_IGRF_up'] = {'long_name':'IGRF Magnetic Field (Up)', 'units':'nT'}
meta['B_IGRF_west'] = {'long_name':'IGRF Magnetic Field (West)', 'units':'nT'}
meta['latitude'] = {'long_name':'Geographic Latituge', 'units':'degrees'}
meta['longitude'] = {'long_name':'Geographic Longitude', 'units':'degrees'}
meta['altitude'] = {'long_name':'Altitude', 'units':'km'}
meta['dB_zon'] = {'long_name':'Delta Magnetic Field - Zonal', 'units':'nT'}
meta['dB_mer'] = {'long_name':'Delta Magnetic Field - Meridional ', 'units':'nT'}
meta['dB_par'] = {'long_name':'Delta Magnetic Field - Parallel ', 'units':'nT'}
                                
def load(fnames, tag=None):
    if len(fnames) <= 0 :
        return pysat.DataFrame(None), None
    else:
         cdf = pycdf.CDF(fnames[0])
         data = {}
         for key in cdf.iterkeys():
             data[key] = cdf[key][...]
         epoch = data.pop['Epoch']
	 data = pysat.DataFrame(data, index=pds.to_datetime(epoch, unit='s'))
	 return data, meta.copy()

def download(date_array, tag, data_path=None, user=None, password=None):
    """
    download vefi 1_second magnetic field data, layout consistent with pysat

    start and stop should be datetimes
    """
    import os
    import ftplib

    ftp = ftplib.FTP('cdaweb.gsfc.nasa.gov')   # connect to host, default port
    ftp.login()               # user anonymous, passwd anonymous@
    
    if tag == 'dc_b':
        ftp.cwd('/pub/data/cnofs/vefi/bfield_1sec')
    
        for date in date_array:
            fname = '{year1:4d}/cnofs_vefi_bfield_1sec_{year2:4d}{month:02d}{day:02d}_v05.cdf'
            fname = fname.format(year1=date.year, year2=date.year, month=date.month, day=date.day)
            local_fname = 'cnofs_vefi_bfield_1sec_{year:4d}{month:02d}{day:02d}_v05.cdf'.format(
                    year=date.year, month=date.month, day=date.day)
            saved_fname = os.path.join(data_path,local_fname) 
            try:
                print 'Downloading file for '+date.strftime('%D')
                ftp.retrbinary('RETR '+fname, open(saved_fname,'w').write)
            except ftplib.error_perm as exception:
                if exception[0][0:3] != '550':
                    raise
                else:
                    print 'File not available for '+ date.strfrtime('%D')



import os
import pandas as pds

from pysat import DataFrame, Series

class Meta(object):
    """
    Stores metadata for an associated pysat Satellite instance. 
    Similar to CF-1.6 data standard info.
    
    inputs:
        metadata: pandas DataFrame indexed by variable name that contains
         atleast the standard_name (name), units, and long_name for the data 
         stored in the associated pysat Satellite object.
    """
    def __init__(self, metadata=None):
        self.replace(metadata=metadata)    

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False    

    def copy(self):
        from copy import deepcopy as deepcopy
        """Deep copy of the meta object."""
        return deepcopy(self) 
               
    def __setitem__(self, name, value):
        """add metadata information"""
        if isinstance(value,dict):
            if 'units' not in value.keys():
                raise ValueError('Must provide "units" in input dictionary.')
            if 'long_name' not in value.keys():
                value['long_name'] = name  
                
            if hasattr(value[value.keys()[0]], '__iter__' ):
                # an iterable of things         
                for key in value.keys():
                    if len(name) != len(value[key]):
                        raise ValueError('Length of names and inputs must be equal.')
                new = DataFrame(value, index=name)
                self.data = self.data.append(new) 
            else:
                new = Series(value)
                self.data.ix[name] = new
                
        if isinstance(value, Series):
            self.data.ix[name] = value
        
        
    def __getitem__(self,key):
        return self.data.ix[key]
        """return metadata"""
    
    #def restore(self):
    #    """Restore metadata from saved metadata"""
    #    self.data = self._orig_data
        
    def replace(self, metadata=None):
        """Replace metadata and saved metadata with input data.
        
        inputs:
            metadata: pandas DataFrame indexed by variable name that contains
            standard_name(name), units, and long_name for the data stored in 
            the associated pysat Satellite object.
        """
        if metadata is not None:
            if isinstance(metadata, DataFrame):
                self.data = metadata
            else:
                raise ValueError("Input must be a pandas DataFrame type."+
                            "See other constructors for alternate inputs.")
            
        else:
            self.data = DataFrame(None, columns=['long_name', 'units'])
        #self._orig_data = self.data.copy()
        
    @classmethod
    def from_csv(cls, name=None, col_names=None, sep=None, **kwargs):
        """
        Create satellite metadata object from csv
        """
        import pysat
        req_names = ['name','long_name','units']
        if col_names is None:
            col_names = req_names
        elif not all([i in col_names for i in req_names]):
            raise ValueError('col_names must include name, long_name, units.')
            
        if sep is None:
            sep = ','
        
        if name is None:
            raise ValueError('Must supply an instrument name or file path.')
        elif not isinstance(name, str):
            raise ValueError('keyword name must be related to a string')               
        else:
            if not os.path.isfile(name):
                # Not a real file, assume input is a pysat instrument name
                # and look in the standard pysat location.
                test =  os.path.join(pysat.__path__,'instruments',name+'_meta.txt')
                #print test
                if os.path.isfile(test):
                    name = test
                else:
                    #trying to form an absolute path for success
                    test = os.path.abspath(name)
                    #print test
                    if not os.path.isfile(test):
                        raise ValueError("Unable to create valid file path.")
                    else:
                        #success
                        name = test
                                     
        mdata = pds.read_csv(name, names=col_names, sep=sep, **kwargs) 
        if not mdata.empty:
            #make sure the data name is the index
            mdata.index = mdata['name']
            return cls(metadata=mdata)
        else:
            raise ValueError('Unable to retrieve information from ' + name)
        
    
    def from_nc():
        """load metadata from netCDF"""
        pass
    
    def from_dict():
        """load metadata from dict of items/list types"""
        pass



# -*- coding: utf-8 -*-
"""
Supports downloading, loading, and cleaning measurements from the
Ion Velocity Meter onboard the Communication and Navigation Outage Forecasting
System (C/NOFS) satellite.
"""
import sys
import os

import pandas as pds
import numpy as np

import pysat
from spacepy import pycdf


def list_files(tag=None, data_path=None):
    """Return a Pandas Series of every file for chosen satellite data"""

    if tag is not None:
        if tag == '':
            return pysat.Files.from_os(data_path=data_path, 
                format_str='cnofs_cindi_ivm_500ms_{year:4d}{month:02d}{day:02d}_v01.cdf')
        else:
            raise ValueError('Unrecognized tag name for C/NOFS IVM')                  
    else:
        raise ValueError ('A tag name must be passed to the loading routine for C/NOFS')           
                
           
def load(fnames, tag=None):
    if len(fnames) <= 0 :
        return pysat.DataFrame(None), pysat.Meta(None)
    else:
         cdf = pycdf.CDF(fnames[0])
         data = {}
         for key in cdf.iterkeys():
             if key not in ['ECISC_matrix','ECISC_index', 'ECISC_index1',
                            'LVLHSC_matrix','LVLHSC_index', 'LVLHSC_index1']:
                data[key] = cdf[key][...]
         # matrices have storage issues (double split intwo two floats),
         # defer issue and drop for now
         epoch = data.pop('Epoch')
	 data = pysat.DataFrame(data, index=epoch)
	 return data, pysat.Meta(None)

def default(ivm):

    ivm.sample_rate = 1.0 if ivm.date >= pysat.datetime(2010, 7, 29) else 2.0
   
            
def clean(self):

    # cleans cindi data
    if self.clean_level == 'clean':
        #choose areas below 550km
        #self.data = self.data[self.data.alt <= 550]
        idx, = np.where(self.data.altitude <= 550)
        self.data = self.data.ix[idx,:]
    
    # make sure all -999999 values are NaN
    #self.data.replace(-999999., np.nan)
    idx, idy, = np.where(self.data == -999999.)
    self.data.ix[idx,idy] = np.nan

   
    if ((self.clean_level == 'clean') | (self.clean_level == 'dusty')):
	try:
	    #self.data = self.data[(abs(self.data.iv_mer) < 10000.)]
	    idx, = np.where(np.abs(self.data.ionVelmeridional) < 10000.)
	    self.data = self.data.ix[idx,:]
        except AttributeError:
            pass
        #take out all values where RPA data quality is > 3
        #self.data = self.data[self.data.rpa_flag <= 3]
        idx, = np.where(self.data.RPAflag <= 3)
        self.data = self.data.ix[idx,:]

        #enforce minimum RPA density if RPA flag eqal to 3
        o_dens = self.data.ionDensity*self.data.ion1fraction
        #self.data = self.data[-((o_dens < 3.E4) & (self.data.rpa_flag==3))]
        idx, = np.where(-((o_dens < 3.E4) & (self.data.RPAflag==3)))
        self.data = self.data.ix[idx,:]
        
        #IDM quality flags
        self.data = self.data[ (self.data.driftMeterflag>= 90) & (self.data.driftMeterflag % 10 < 1) ]
        idx, = np.where((self.data.driftMeterflag>= 90) & (self.data.driftMeterflag % 10 < 1))
        self.data = self.data.ix[idx,:]

    # sometimes yrdoyentry is 0, remove them
    #self.data = self.data[self.data.yrdoy != 0]
    idx, = np.where(self.data.yrdoy != 0)
    self.data = self.data.ix[idx,:]

    # basic quality check on drifts and don't let UTS go above 86400.
    #self.data = self.data[ (self.data.uts <= 86400.)]
    idx, = np.where(self.data.time <= 86400.)
    self.data = self.data.ix[idx,:]
    
    #make sure MLT is between 0 and 24
    #self.data = self.data[(self['mlt'] >= 0.) & (self['mlt'] <= 24.)]
    idx, = np.where((self.data.mlt >= 0) & (self.data.mlt <= 24.))
    self.data = self.data.ix[idx,:]


    return  
    
def download(date_array, tag, data_path, user=None, password=None):
    """
    download IVM data consistent with pysat

    """
    import ftplib
    from ftplib import FTP
    import sys
    ftp = FTP('cdaweb.gsfc.nasa.gov')   # connect to host, default port
    ftp.login()               # user anonymous, passwd anonymous@
    ftp.cwd('/pub/data/cnofs/cindi/ivm_500ms_cdf')


    for date in date_array:
        fname = '{year1:4d}/cnofs_cindi_ivm_500ms_{year2:4d}{month:02d}{day:02d}_v01.cdf'
        fname = fname.format(year1=date.year, year2=date.year, 
                                month=date.month, day=date.day)
        local_fname = 'cnofs_cindi_ivm_500ms_{year:4d}{month:02d}{day:02d}_v01.cdf'.format(
                        year=date.year, month=date.month, day=date.day)
        saved_fname = os.path.join(data_path,local_fname) 
        try:
            print 'Downloading file for '+date.strftime('%D')
            sys.stdout.flush()
            ftp.retrbinary('RETR '+fname, open(saved_fname,'w').write)
        except ftplib.error_perm as exception:
            print exception[0][0:3]
            if exception[0][0:3] != '550':
                raise
            else:
                print 'File not available for '+date.strftime('%D')


            



import numpy as np
import pandas as pds
from pysat import Series, DataFrame

class orbits(object):
    
    
#####################
### The following routines allow the user to select a particular orbit
### from a satellite. It also inlcudes an orbit iterator that can span day breaks,
### freeing the user from this data management task.
#####################

    def __init__(self, sat=None, index=None, kind=None, period=None):
        #create null arrays for storing orbit info
        if sat is None:
            raise ValueError('Must provide a pysat instrument object when initializing '+ 
                                'orbits class.')
        else:
            #self.sat = weakref.proxy(sat)
            self.sat = sat
            
        if kind is None:
            kind = 'local time'
        else:
            kind = kind.lower()
        
        if period is None:
            period = np.timedelta64(97, 'm')
        self.orbit_period = period

        if (kind == 'local time') or (kind == 'lt'):
            self._detBreaks = self._equaBreaks
        elif (kind == 'longitude') or (kind == 'long'):
            self._detBreaks = self._longBreaks
        elif kind == 'polar':
            self._detBreaks = self._polarBreaks
        else:
            raise ValueError('Unknown kind of orbit requested.')
       
        self._orbit_breaks = []
        self.num = []
        self.current = 0
        self.orbit_index = index

            
    def __getitem__(self, key): 
        """ Enable convenience notation for loading orbit into parent object.
        
        """
        
        self.load(key)

    def _reset(self):
        #create null arrays for storing orbit info
        self._orbit_breaks = []
        self.num = None
        self.current = 0

    def _calcOrbits(self):
        """Prepares data structure for breaking data into orbits. Not intended for end user."""
        #if the breaks between orbit have not been defined, define them
        #also, store the data so that grabbing different orbits does not
        #require reloads of whole dataset
        if self._orbit_breaks == []:
                #determine orbit breaks
                self._detBreaks()
                #store a copy of data
                self._fullDayData = self.sat.data.copy()
                #set current orbit counter to zero (default)
                self.current = 0

    def _equaBreaks(self):
        """Determine where breaks in an equatorial satellite orbit occur.
        
        Looks for negative gradients in local time (or longitude) as well as 
        breaks in UT.

	"""
     
        if self.orbit_index is None:
            try:
    	        self.sat['slt']
		self.orbit_index = 'slt'
            except ValueError:
                try:
                    self.sat['mlt']
                    self.orbit_index = 'mlt'
                except ValueError:
                    try:
                        self.sat['glong']
			self.orbit_index = 'glong'
                    except ValueError:
                        raise ValueError('Unable to find a valid index (slt/mlt/glong) for determining orbits.')
        else:
            try:
                self.sat[self.orbit_index]
            except ValueError:
                raise ValueError('Provided orbit index does not exist in loaded data')

        #define derivative function here
        #deriv = lambda x: (x[1] - x[0])/2
        lt_diff = self.sat[self.orbit_index].diff()
        ut_vals = Series(self.sat.data.index)
        ut_diff = ut_vals.diff()
        #locations where derivative is less than 0
        #or, look for breaks because the length of time between samples is too large
        #deriv of datetime index is in naneseconds
        ind, = np.where((lt_diff < -0.1) )
        if len(ind) > 0:
            ind = np.hstack((ind, np.array([ len(self.sat[self.orbit_index]) ]) ))
            #look at distance between breaks
            dist = ind[1:] - ind[0:-1]
            #only keep orbit breaks with a distance greater than 1
            #done for robustness
            if len(ind) > 1:
                if min(dist)==1:
                    print 'There are orbit breaks right next to each other'
                ind = ind[dist>1]
                
            #check for large positive gradients around the break that would
            #suggest not a true orbit break, but rather bad orbit_index values
            new_ind = []
            for idx in ind:
                tidx, = np.where(lt_diff[idx-5:idx+6] > 0.1)

		if len(tidx) != 0:
  		    for tidx in tidx:
      		        #look at time change vs local time change
      		        if ut_diff[idx-5:idx+6].iloc[tidx] < lt_diff[idx-5:idx+6].iloc[tidx]/24.*self.orbit_period :
      		            #change in ut inconsistent with change in local time
      		            #increases in local time require a change in ut	
      		            pass
      		        #print 'Fake Orbit Break.'
      		        else:
      		            new_ind.append(idx)	
      		            break    
		else:
		    new_ind.append(idx)

	    ind = np.array(new_ind)
	    
	    
	    #while this works correctly, when it works correctly it doesn't give
	    #what I need
            ##look at the time change vs lt change for the negative gradients
            #new_ind = []
            #for idx in ind:
            #    diff = lt_diff.iloc[idx]
            #    print 'Checking negative gradient for validity at UT', self.sat.data.index[idx]
            #    print 'MLT values around break ', self.sat[self.orbit_index][idx-3:idx+4]
            #    print 'UT Diff ', ut_diff.iloc[idx]
            #    print 'LT Diff ', diff
            #    print 'Equivalent UT Drift ', np.abs(1. + diff/24.)*self.orbit_period
            #    if ut_diff.iloc[idx] < np.abs(1. + diff/24.)*self.orbit_period:
            #        #for almost 24 hours of local time change, there should be no
            #        # time change (<= 1 second)
            #        print "Negative gradient in MLT doesn't have corresponding UT change"
            #        print "Not a true orbit break (likely)"
            #        pass
            #    else:
            #        new_ind.append(idx)  
            #ind = new_ind        

        
        ##now, look for breaks because the length of time between samples is too large, thus there is no break in slt/mlt/etc
        #ut_val = pds.rolling_apply(self.sat.data.index.values,2,deriv)
        #ut_val has time differences in nanseconds, so orbit_period expressed in nano here
        ut_ind, = np.where((ut_diff > self.orbit_period) | ((ut_diff/self.orbit_period > (np.abs(lt_diff/24.)) ) & (ut_diff/self.orbit_period > 0.95) ))#  & lt_diff.notnull() ))# & (lt_diff != 0)  ) )   #added the or and check after or on 10/20/2014
        if len(ut_ind) > 0:
            #print len(ut_ind), len(ut_diff)
            ind = np.hstack((ind,ut_ind))
            ind = np.sort(ind)
            ind = np.unique(ind)
            print 'Time Gap'
            #print ut_diff[1]/self.orbit_period*24., lt_diff[1], ut_diff[1], 0.97*self.orbit_period
            #print ut_diff.iloc[ut_ind], lt_diff.iloc[ut_ind], np.abs(lt_diff.iloc[ut_ind]/24.), ut_diff[ut_ind]/self.orbit_period
            #print np.abs(lt_diff.iloc[ut_ind]/24.*self.orbit_period)

	#now that most problems in orbits should have been caught, look at 
	#the time difference between orbits
	orbit_ut_diff = ut_vals[ind].diff()
	orbit_lt_diff = self.sat[self.orbit_index][ind].diff()
	#print 'Differences in orbit time before last filter'
	#print ut_vals[ind]
	#print orbit_ut_diff
	#print orbit_lt_diff
	#
	#print 'Scaled Time Difference'
	#print orbit_ut_diff - orbit_lt_diff.values/24.*self.orbit_period
	idx, = np.where((orbit_ut_diff/self.orbit_period - orbit_lt_diff.values/24.) > 0.97)
	idx = np.hstack((0,idx))
	if len(ind) > 0:
	    ind = ind[idx]
	#print 'Differences in orbit time after last filter'
	#print ut_vals[ind].diff()


        # number of orbits
        num_orbits = len(ind)+1
        # create orbitbreak index
        ind=np.hstack((np.array([0]),ind))
        # set index of orbit breaks
        self._orbit_breaks = ind
        # set number of orbits for the day
        self.num = num_orbits


    def _longBreaks(self):
        """
        Determine where breaks in a satellite orbit occur, based upon
        changes in longitude.
        
        Looks for negative gradients in longitude as well as 
        breaks in UT.

	"""
     
        if self.orbit_index is None:
            try:
                self.sat['glong']
		self.orbit_index = 'glong'
            except ValueError:
                raise ValueError(''.join(('Unable to find a valid index',
                    'for determining orbits. Provide one in orbit_index')))
        else:
            try:
                self.sat[self.orbit_index]
            except ValueError:
                raise ValueError('Provided orbit index does not exist in loaded data')

        lt_diff = self.sat[self.orbit_index].diff()
        ut_vals = Series(self.sat.data.index)
        ut_diff = ut_vals.diff()
        # locations where derivative is less than 0
        # or, look for breaks because the length of time between samples is too large
        # deriv of datetime index is in naneseconds
        ind, = np.where((lt_diff < -0.1) )
        if len(ind) > 0:
            ind = np.hstack((ind, np.array([ len(self.sat[self.orbit_index]) ]) ))
            # look at distance between breaks
            dist = ind[1:] - ind[0:-1]
            # only keep orbit breaks with a distance greater than 1
            # done for robustness, though this could rarely introduce a problem
            if len(ind) > 1:
                if min(dist)==1:
                    print 'There are orbit breaks right next to each other'
                ind = ind[dist>1]
                
            # check for large positive gradients around the break that would
            # suggest not a true orbit break, but rather bad orbit_index values
            new_ind = []
            for idx in ind:
                tidx, = np.where(lt_diff[idx-5:idx+6] > 0.1)

		if len(tidx) != 0:
  		    for tidx in tidx:
      		        # look at time change vs longitude change
      		        if ut_diff[idx-5:idx+6].iloc[tidx] < lt_diff[idx-5:idx+6].iloc[tidx]/360.*self.orbit_period :
      		            # change in ut inconsistent with change in local time
      		            # increases in longitude require a change in ut	
      		            pass
      		        #print 'Fake Orbit Break.'
      		        else:
      		            new_ind.append(idx)	
      		            break    
		else:
		    new_ind.append(idx)

	    ind = np.array(new_ind)
	    
        
        # now, look for breaks because the length of time between samples is too large, thus there is no break in slt/mlt/etc
        ut_ind, = np.where(
                        (ut_diff > self.orbit_period) | 
                        ( (ut_diff/self.orbit_period > (np.abs(lt_diff/360.))) & 
                        (ut_diff/self.orbit_period > 0.95) )
                        )
                    #  & lt_diff.notnull() ))# & (lt_diff != 0)  ) )   #added the or and check after or on 10/20/2014
        
        if len(ut_ind) > 0:
            ind = np.hstack((ind,ut_ind))
            ind = np.sort(ind)
            ind = np.unique(ind)
            print 'Time Gap'

	# now that most problems in orbits should have been caught, look at 
	# the time difference between orbits
	orbit_ut_diff = ut_vals[ind].diff()
	orbit_lt_diff = self.sat[self.orbit_index][ind].diff()

	idx, = np.where((orbit_ut_diff/self.orbit_period - orbit_lt_diff.values/360.) > 0.97)
	idx = np.hstack((0,idx))
	if len(ind) > 0:
	    ind = ind[idx]

        # number of orbits
        num_orbits = len(ind)+1
        # create orbitbreak index
        ind=np.hstack((np.array([0]),ind))
        # set index of orbit breaks
        self._orbit_breaks = ind
        # set number of orbits for the day
        self.num = num_orbits


    def _polarBreaks(self):
        """Determine where breaks in a polar orbiting satellite orbit occur.
        
        Looks for sign changes in latitude (magnetic or geographic) as well as 
        breaks in UT.

	"""
     
        if self.orbit_index is None:
            try:
    	        self.sat['glat']
		self.orbit_index = 'glat'
            except ValueError:
                try:
                    self.sat['mlat']
                    self.orbit_index = 'mlat'
                except ValueError:
                    raise ValueError('Unable to find a valid index (glat/mlat)'+
                                    ' for determining orbits.')
        else:
            try:
                self.sat[self.orbit_index]
            except ValueError:
                raise ValueError('Provided orbit index does not exist in loaded data')

        # define derivative function here
        deriv = lambda x: (x[1] - x[0])/2
        val = self.sat[self.orbit_index] >= 0
        val_change = val[:-1] - val[1:]
        ut_val = pds.rolling_apply(self.sat.data.index.values,2,deriv)
        # locations where derivative is less than 0
        # or, look for breaks because the length of time between samples is too large
        # deriv of datetime index is in naneseconds
        ind, = np.where(val_change == True )
        if len(ind) > 0:
            ind = np.hstack((ind, np.array([ len(self.sat[self.orbit_index]) ]) ))
            # look at distance between breaks
            dist = ind[1:] - ind[0:-1]
            # only keep orbit breaks with a distance greater than 1
            # done for robustness
            if len(ind) > 1:
                if min(dist)==1:
                    print 'There are orbit breaks right next to each other'
                ind = ind[dist>1]
        # now, look for breaks because the length of time between samples is too large, thus there is no break in slt/mlt/etc
        ut_val = pds.rolling_apply(self.sat.data.index.values,2,deriv)
        ut_ind, = np.where(ut_val > self.orbit_period.total_seconds()*10**9 )     #used to be 2900, dunno why. 5900 corresponds to seconds over 97 minutes (1 orbit)
        if len(ut_ind) > 0:
            ind = np.hstack((ind,ut_ind))
            ind = np.sort(ind)
            ind = np.unique(ind)
            print 'Time Gap'
        # number of orbits
        num_orbits = len(ind)+1
        # create orbitbreak index
        ind=np.hstack((np.array([0]),ind))
        # set index of orbit breaks
        self._orbit_breaks = ind
        # set number of orbits for the day
        self.num = num_orbits



    def _getBasicOrbit(self, orbit=None):
        """Load a particular orbit into .data for loaded day.

        Keywords:
            orbit = orbit number, 1 indexed, negative indexes allowed, -1 last orbit

        Note:
            A day of data must be loaded before this routine functions properly.
            If the last orbit of the day is requested, it will NOT automatically be
            padded with data from the next day."""
        #ensure data exists
        if len(self.sat.data) > 0:
            #ensure proper orbit metadata present
            self._calcOrbits()

            #ensure user is requesting a particular orbit
            if orbit is not None:
                #pull out requested orbit
                if orbit == -1:
                    #load orbit data into data
                    self.sat.data = self._fullDayData[self._orbit_breaks[self.num+orbit]:]
                    self.current = self.num+orbit+1
                elif ((orbit < 0) & (orbit >= -self.num)):
                    #load orbit data into data
                    self.sat.data = self._fullDayData[self._orbit_breaks[self.num+orbit]:self._orbit_breaks[self.num+orbit+1]]
                    self.current = self.num+orbit+1                    
                elif (orbit < self.num) & (orbit != 0):
                    #load orbit data into data
                    self.sat.data = self._fullDayData[self._orbit_breaks[orbit-1]:self._orbit_breaks[orbit]]
                    self.current = orbit
                elif orbit==self.num:
                    self.sat.data = self._fullDayData[self._orbit_breaks[orbit-1]:]
                    self.current = orbit #recent addition, wondering why it wasn't there before, could just be a bug
                elif orbit == 0:
		    raise ValueError('Orbits start at 1, 0 not allowed')	
                else:
                    #gone too far
                    self.sat.data = []
                    raise ValueError('Requested an orbit past total orbits for day')
            else:
                raise ValueError('Must set an orbit')

    def load(self, orbit=None):
        """Load a particular orbit into .data for loaded day.

        Keywords:
            orbit = orbit number, 1 indexed

        Note:
            A day of data must be loaded before this routine functions properly.
            If the last orbit of the day is requested, it will automatically be
            padded with data from the next day. The orbit counter will be reset to 1."""
        if len(self.sat.data) > 0: #ensure data exists
            #set up orbit metadata
            self._calcOrbits()
            #ensure user supplied an orbit
            if orbit != None:
                #pull out requested orbit
                if orbit < 0:
                    #negative indexing consistent with numpy, -1 last, -2 second
                    # to last, etc.
                    orbit = self.num+1+orbit

                if orbit == 1:
                    #change from orig copied from _core, didn't look correct.
      #              self._getBasicOrbit(orbit=2)
		    #self.prev()	
		    try:
			true_date = self.sat.date#.copy()

      		        self.sat.prev()
      		        #if and else added becuase of CINDI turn off 6/5/2013,turn on 10/22/2014
      		        #crashed when starting on 10/22/2014
      		        #prev returned empty data
      		    	if len(self.sat.data) >0:
      		    	    self.load(orbit=-1) 
      		    	else:
      		    	    self.sat.next()
      		    	    self._getBasicOrbit(orbit=1)
                        #check that this orbit should end on the current day
                        delta = pds.to_timedelta(true_date-self.sat.data.index[0])
                        #print 'checking if first orbit should land on requested day'
                        #print self.sat.date, self.sat.data.index[0], delta, delta >= self.orbit_period
                        #print delta - self.orbit_period
                        if delta >= self.orbit_period:
                            #the orbit loaded isn't close enough to date
                            #to be the first orbit of the day, move forward
                            self.next()
		    except StopIteration:
		        #print 'going for basic orbit'
		        self._getBasicOrbit(orbit=1)
		        print 'Loaded Orbit:%i' % self.current
                    #check if the first orbit is also the last orbit
                    
                elif orbit==self.num:
                    #we get here if user asks for last orbit
                    #make sure that orbit data goes across daybreak as needed
                    #load previous orbit
                    if self.num != 1:
                        self._getBasicOrbit(self.num-1)
                        self.next()
                    else:
                        #self.current = 1
                        self._getBasicOrbit(orbit=-1)
                    #request next orbit, this will go across day break nicely
                    #self.next()

                elif orbit < self.num:
                    #load orbit data into data
                    self._getBasicOrbit(orbit)
                    print 'Loaded Orbit:%i' % self.current

                else:
                    #gone too far
                    self.sat.data = []
                    raise Exception('Requested an orbit past total orbits for day')
            else:
                raise Exception('Must set an orbit')
        else:
            print 'No data loaded in instrument object to determine orbits.'
            
    def next(self, *arg, **kwarg):
        """Load the next orbit into .data.

        Notes:
            Forms complete orbits across day boundaries."""

        #first, check if data exists
        if len(self.sat.data) > 0:
            #set up orbit metadata
            self._calcOrbits()

             #if current orbit near the last, must be careful
            if self.current == (self.num - 1 ):
                #first, load last orbit data
                self._getBasicOrbit(orbit=-1)
                #should come up with some kind of check if the next day is needed
                load_next = True
                if self.sat._iter_type == 'date':
                    delta = pds.to_timedelta(self.sat.date - self.sat.data.index[-1]) + np.timedelta64(1, 'D') 
                    #print 'Checking if new load needed ', delta < self.orbit_period
                    if delta > self.orbit_period:
                        #don't need to load the next day
                        #everything is already done
                        #print 'Not loading next day to give last orbit'
                        load_next = False
                #need to save this current orbit and load the next day
                if load_next:
                    temp_orbit_data = self.sat.data.copy() 
                    try:
                        #loading next day/file clears orbit breaks info
                        self.sat.next()
   	            #combine this next day orbit with previous last orbit
                        if len(self.sat.data) > 0:
                            self.sat.data = pds.concat([temp_orbit_data[:self.sat.data.index[0] - pds.DateOffset(microseconds=1)], self.sat.data])
                            self._getBasicOrbit(orbit=1)
                        else:
                            self.sat.prev() 
                            self._getBasicOrbit(orbit=-1)
                    except StopIteration:
                        pass
                    del temp_orbit_data
                print 'Loaded Orbit:%i' % self.current  

            elif self.current == (self.num):
                #self._getBasicOrbit(orbit=-1)
                #need to save this current orbit and load the next day
                temp_orbit_data = self.sat.data.copy() 
                #load next day, which clears orbit breaks info
                self.sat.next() 
                #combine this next day orbit with previous last orbit
                if len(self.sat.data) > 0:
                    #check if data padding is really needed, only works when loading by date
                    pad_next = True
                    if self.sat._iter_type == 'date':
                        delta = pds.to_timedelta(self.sat.date-temp_orbit_data.index[-1])
                        #print 'Checking if new load was needed ', (delta < self.orbit_period)
                        if delta > self.orbit_period:
                            #don't need to load the next day
                            pad_next = False                    
                    if pad_next:
                        self.sat.data = pds.concat([temp_orbit_data[:self.sat.data.index[0] - pds.DateOffset(microseconds=1)], self.sat.data])
                        #select second orbit of combined data
                        self._getBasicOrbit(orbit=2)
		    else:
		        self._getBasicOrbit(orbit=1)
                        if self.sat._iter_type == 'date':
                            delta = pds.to_timedelta(self.sat.date + pds.DateOffset(days=1)-self.sat.data.index[0])
                            #print 'Checking if another new load is needed ', delta < self.orbit_period
 			    if delta < self.orbit_period:
 			        #first orbit is also last orbit which spans to next day
  		                self.current = self.num - 1
  		                self.next()		            

                else:
                    #continue loading data until there is some
                    #nextData raises StopIteration when it reaches the end
                    while len(self.sat.data) == 0:
                        self.sat.next()
                    self._getBasicOrbit(orbit=1)

                #print 'Loaded Orbit:%i' % self.current
                del temp_orbit_data
                print 'Loaded Orbit:%i' % self.current  

           #get next orbit
	    elif self.current == 0:
	        #autoloads prev day if needed to form complete orbit
	        self.load(orbit=1)

            #if not close to the last orbit,just pull the next orbit
            elif self.current < (self.num - 1):
                self._getBasicOrbit(orbit=self.current+1)
                print 'Loaded Orbit:%i' % self.current  

            else:
                raise Exception('You ended up where noone should ever be. Talk to someone about this fundamental failure.')
              
        else: #no data
    	    while len(self.sat.data) == 0:
    	        self.sat.next()#raises stopIteration at end of dataset
            self.next()
        
    def prev(self, *arg, **kwarg):
        """Load the next orbit into .data.

        Notes:
            Forms complete orbits across day boundaries."""

        #first, check if data exists
        if len(self.sat.data) > 0:
            #set up orbit metadata
            self._calcOrbits()
            #if not close to the first orbit,just pull the previous orbit
  #          if self.current == self.num:
		#print 'max orbit'
  #              self._getBasicOrbit(orbit=-2)    
                #print 'Loaded Orbit:%i' % self.current

	    if (self.current > 2) & (self.current <= self.num):
                 #load orbit and put it into self.sat.data
                self._getBasicOrbit(orbit=self.current-1)
                #print 'Loaded Orbit:%i' % self.current

            #if current orbit near the last, must be careful
            elif self.current == 2:
                #first, load prev orbit data
                self._getBasicOrbit(orbit=self.current-1)
                #need to save this current orbit and load the next day
                temp_orbit_data = self.sat.data[self.sat.date:]
                #load previous day, which clears orbit breaks info

                try:
                    self.sat.prev() 
                    #combine this next day orbit with previous last orbit
                    if len(self.sat.data) > 0:
                        self.sat.data = pds.concat([self.sat.data, temp_orbit_data])
                        #select first orbit of combined data
                        self._getBasicOrbit(orbit=-1)
                    else:
                        self.sat.next()
                        self._getBasicOrbit(orbit=1)
                except StopIteration:
                    #if loading the first orbit, of first day of data, you'll
                    #end up here as the attempt to make a full orbit will 
                    #move the date backwards, and StopIteration is made.
                    #everything is already ok, just move along
                    pass

                #print 'Loaded Orbit:%i' % self.current
                #clear temporary variable
                del temp_orbit_data

            elif self.current < 2:
                #first, load next orbit data
                self._getBasicOrbit(orbit=1)
                #need to save this current orbit and load the next day
                temp_orbit_data = self.sat.data[self.sat.date:]
                #load previous day, which clears orbit breaks info
                self.sat.prev()
                #combine this next day orbit with previous last orbit
                if len(self.sat.data) > 0:
                    self.sat.data = pds.concat([self.sat.data, temp_orbit_data])
                    #select first orbit of combined data
                    self._getBasicOrbit(orbit=-2)

                #print 'Loaded Orbit:%i' % self.current

                #clear temporary variable
                del temp_orbit_data

            else:
                raise Exception('You ended up where noone should ever be. Talk to someone about this fundamental failure.')
            print 'Loaded Orbit:%i' % self.current 
        else:
           #no data
    	    while len(self.sat.data) == 0:
    	        self.sat.prev()#raises stopIteration at end of dataset
            self.prev()    
            
    def __iter__(self):
        '''Support iteration by orbit. Uses bounds of Satellite object.
        	
	Limits of iteration set by calling setBounds method.	
        '''
        while True:
            self.next()
            yield self.sat
            
   